@@ -45,21 +45,27 @@ comparison function when necessary. But most of the time treating the files
 as binary does the job.
 -}
 
+{-# LANGUAGE OverloadedStrings #-}
 module Test.Tasty.Golden
   ( goldenVsFile
   , goldenVsString
   , goldenVsFileDiff
   , goldenVsStringDiff
+  , SizeCutoff(..)
   , writeBinaryFile
   , findByExtension
+  , createDirectoriesAndWriteFile
   )
   where
 
-import Test.Tasty.Providers
+import Test.Tasty
 import Test.Tasty.Golden.Advanced
+import Test.Tasty.Golden.Internal
 import Text.Printf
-import qualified Data.ByteString as BS
 import qualified Data.ByteString.Lazy as LBS
+import Data.Monoid
+import qualified Data.Text.Lazy as LT
+import qualified Data.Text.Lazy.Encoding as LT
 import System.IO
 import System.IO.Temp
 import System.Process
@@ -68,11 +74,10 @@ import System.FilePath
 import System.Directory
 import Control.Exception
 import Control.Monad
-import Control.Applicative
-import Control.DeepSeq
 import qualified Data.Set as Set
 
--- | Compare a given file contents against the golden file contents
+-- | Compare the output file's contents against the golden file's contents
+-- after the given action has created the output file.
 goldenVsFile
   :: TestName -- ^ test name
   -> FilePath -- ^ path to the «golden» file (the file that contains correct output)
@@ -82,32 +87,34 @@ goldenVsFile
 goldenVsFile name ref new act =
   goldenTest
     name
-    (BS.readFile ref)
-    (act >> BS.readFile new)
+    (readFileStrict ref)
+    (act >> readFileStrict new)
     cmp
     upd
   where
   cmp = simpleCmp $ printf "Files '%s' and '%s' differ" ref new
-  upd = BS.writeFile ref
+  upd = createDirectoriesAndWriteFile ref
 
--- | Compare a given string against the golden file contents
+-- | Compare a given string against the golden file's contents.
 goldenVsString
   :: TestName -- ^ test name
   -> FilePath -- ^ path to the «golden» file (the file that contains correct output)
   -> IO LBS.ByteString -- ^ action that returns a string
   -> TestTree -- ^ the test verifies that the returned string is the same as the golden file contents
 goldenVsString name ref act =
+  askOption $ \sizeCutoff ->
   goldenTest
     name
-    (BS.readFile ref)
-    (LBS.toStrict <$> act)
-    cmp
+    (readFileStrict ref)
+    act
+    (cmp sizeCutoff)
     upd
   where
-  cmp x y = simpleCmp msg x y
+  cmp sizeCutoff x y = simpleCmp msg x y
     where
-    msg = printf "Test output was different from '%s'. It was: %s" ref (show y)
-  upd = BS.writeFile ref
+    msg = printf "Test output was different from '%s'. It was:\n" ref <>
+      unpackUtf8 (truncateLargeOutput sizeCutoff y)
+  upd = createDirectoriesAndWriteFile ref
 
 simpleCmp :: Eq a => String -> a -> a -> IO (Maybe String)
 simpleCmp e x y =
@@ -128,27 +135,28 @@ goldenVsFileDiff
   -> IO ()    -- ^ action that produces the output file
   -> TestTree
 goldenVsFileDiff name cmdf ref new act =
+  askOption $ \sizeCutoff ->
   goldenTest
     name
     (return ())
     act
-    cmp
+    (cmp sizeCutoff)
     upd
   where
   cmd = cmdf ref new
-  cmp _ _ | null cmd = error "goldenVsFileDiff: empty command line"
-  cmp _ _ = do
+  cmp sizeCutoff _ _
+    | null cmd = error "goldenVsFileDiff: empty command line"
+    | otherwise = do
     (_, Just sout, _, pid) <- createProcess (proc (head cmd) (tail cmd)) { std_out = CreatePipe }
     -- strictly read the whole output, so that the process can terminate
-    out <- hGetContents sout
-    evaluate . rnf $ out
+    out <- hGetContentsStrict sout
 
     r <- waitForProcess pid
     return $ case r of
       ExitSuccess -> Nothing
-      _ -> Just out
+      _ -> Just . unpackUtf8 . truncateLargeOutput sizeCutoff $ out
 
-  upd _ = BS.readFile new >>= BS.writeFile ref
+  upd _ = readFileStrict new >>= createDirectoriesAndWriteFile ref
 
 -- | Same as 'goldenVsString', but invokes an external diff command.
 goldenVsStringDiff
@@ -164,18 +172,19 @@ goldenVsStringDiff
   -> IO LBS.ByteString -- ^ action that returns a string
   -> TestTree
 goldenVsStringDiff name cmdf ref act =
+  askOption $ \sizeCutoff ->
   goldenTest
     name
-    (BS.readFile ref)
-    (LBS.toStrict <$> act)
-    cmp
+    (readFileStrict ref)
+    (act)
+    (cmp sizeCutoff)
     upd
   where
-  template = takeFileName ref <.> "actual"
-  cmp _ actBS = withSystemTempFile template $ \tmpFile tmpHandle -> do
+  template = takeBaseName ref <.> "actual"
+  cmp sizeCutoff _ actBS = withSystemTempFile template $ \tmpFile tmpHandle -> do
 
     -- Write act output to temporary ("new") file
-    BS.hPut tmpHandle actBS >> hFlush tmpHandle
+    LBS.hPut tmpHandle actBS >> hFlush tmpHandle
 
     let cmd = cmdf ref tmpFile
 
@@ -183,17 +192,27 @@ goldenVsStringDiff name cmdf ref act =
 
     (_, Just sout, _, pid) <- createProcess (proc (head cmd) (tail cmd)) { std_out = CreatePipe }
     -- strictly read the whole output, so that the process can terminate
-    out <- hGetContents sout
-    evaluate . rnf $ out
+    out <- hGetContentsStrict sout
 
     r <- waitForProcess pid
     return $ case r of
       ExitSuccess -> Nothing
-      _ -> Just (printf "Test output was different from '%s'. Output of %s:\n%s" ref (show cmd) out)
+      _ -> Just (printf "Test output was different from '%s'. Output of %s:\n" ref (show cmd) <> unpackUtf8 (truncateLargeOutput sizeCutoff out))
 
-  upd = BS.writeFile ref
+  upd = createDirectoriesAndWriteFile ref
 
--- | Like 'writeFile', but uses binary mode
+truncateLargeOutput
+  :: SizeCutoff
+  -> LBS.ByteString
+  -> LBS.ByteString
+truncateLargeOutput (SizeCutoff n) str =
+  if LBS.length str <= n
+    then str
+    else
+      LBS.take n str <> "<truncated>" <>
+      "\nUse --accept or increase --size-cutoff to see full output."
+
+-- | Like 'writeFile', but uses binary mode.
 writeBinaryFile :: FilePath -> String -> IO ()
 writeBinaryFile f txt = withBinaryFile f WriteMode (\hdl -> hPutStr hdl txt)
 
@@ -237,3 +256,40 @@ findByExtension extsList = go where
             if takeExtension path `Set.member` exts
               then [path]
               else []
+
+-- | Like 'BS.writeFile', but also create parent directories if they are
+-- missing.
+createDirectoriesAndWriteFile
+  :: FilePath
+  -> LBS.ByteString
+  -> IO ()
+createDirectoriesAndWriteFile path bs = do
+  let dir = takeDirectory path
+  createDirectoryIfMissing
+    True -- create parents too
+    dir
+  LBS.writeFile path bs
+
+-- | Force the evaluation of a lazily-produced bytestring.
+--
+-- This is important to close the file handles.
+--
+-- See <https://ro-che.info/articles/2015-05-28-force-list>.
+forceLbs :: LBS.ByteString -> ()
+forceLbs = LBS.foldr seq ()
+
+readFileStrict :: FilePath -> IO LBS.ByteString
+readFileStrict path = do
+  s <- LBS.readFile path
+  evaluate $ forceLbs s
+  return s
+
+hGetContentsStrict :: Handle -> IO LBS.ByteString
+hGetContentsStrict h = do
+  hSetBinaryMode h True
+  s <- LBS.hGetContents h
+  evaluate $ forceLbs s
+  return s
+
+unpackUtf8 :: LBS.ByteString -> String
+unpackUtf8 = LT.unpack . LT.decodeUtf8
