@@ -1,11 +1,7 @@
 {-# LANGUAGE RankNTypes #-}
 module Test.Tasty.Golden.Advanced
   ( -- * The main function
-    goldenTest,
-
-    -- * ValueGetter monad
-    ValueGetter(..),
-    vgReadFile
+    goldenTest
   )
 where
 
@@ -15,8 +11,15 @@ import Test.Tasty.Golden.Internal
 -- | A very general testing function.
 goldenTest
   :: TestName -- ^ test name
-  -> (forall r . ValueGetter r a) -- ^ get the golden correct value
-  -> (forall r . ValueGetter r a) -- ^ get the tested value
+  -> (IO a)
+    -- ^ get the golden correct value
+    --
+    -- Note that this action may be followed by the update function call.
+    --
+    -- Therefore, this action *should avoid* reading the file lazily;
+    -- otherwise, the file may remain half-open and the update action will
+    -- fail.
+  -> (IO a) -- ^ get the tested value
   -> (a -> a -> IO (Maybe String))
     -- ^ comparison function.
     --
@@ -26,6 +29,7 @@ goldenTest
     --
     -- The function may use 'IO', for example, to launch an external @diff@
     -- command.
-  -> (a -> IO ()) -- ^ update the golden file
+  -> (a -> IO ())
+    -- ^ update the golden file
   -> TestTree
 goldenTest t golden test cmp upd = singleTest t $ Golden golden test cmp upd
